"""Encoder-decoder transformer layers for self/cross attention."""

from copy import deepcopy

import torch
from torch import nn
import torch.nn.functional as F

from .modules import PositionEmbeddingLearned
from .deformable_attention import MSDeformAttnPC
import ipdb
st = ipdb.set_trace


def _get_clones(module, N):
    return nn.ModuleList([deepcopy(module) for i in range(N)])


class CrossAttentionLayer(nn.Module):
    """Cross-attention between language and vision."""

    def __init__(self, d_model=256, dropout=0.1, n_heads=8):
        """Initialize layers, d_model is the encoder dimension."""
        super().__init__()

        # Cross attention from lang to vision
        self.cross_lv = nn.MultiheadAttention(
            d_model, n_heads, dropout=dropout
        )
        self.dropout_lv = nn.Dropout(dropout)
        self.norm_lv = nn.LayerNorm(d_model)
        self.ffn_lv = nn.Sequential(
            nn.Linear(d_model, 1024),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(1024, d_model),
            nn.Dropout(dropout)
        )
        self.norm_lv2 = nn.LayerNorm(d_model)

        # Cross attention from vision to lang
        self.cross_vl = deepcopy(self.cross_lv)
        self.dropout_vl = nn.Dropout(dropout)
        self.norm_vl = nn.LayerNorm(d_model)
        self.ffn_vl = deepcopy(self.ffn_lv)
        self.norm_vl2 = nn.LayerNorm(d_model)

    def forward(self, vis_feats, vis_key_padding_mask, text_feats,
                text_key_padding_mask, pos_feats):
        """Forward pass, vis/pos_feats (B, V, F), lang_feats (B, L, F)."""
        # produce key, query, value for image
        qv = kv = vv = vis_feats
        qv = qv + pos_feats  # add pos. feats only on query

        # produce key, query, value for text
        qt = kt = vt = text_feats

        # cross attend language to vision
        text_feats2 = self.cross_lv(
            query=qt.transpose(0, 1),
            key=kv.transpose(0, 1),
            value=vv.transpose(0, 1),
            attn_mask=None,
            key_padding_mask=vis_key_padding_mask  # (B, V)
        )[0].transpose(0, 1)
        text_feats = text_feats + self.dropout_lv(text_feats2)
        text_feats = self.norm_lv(text_feats)
        text_feats = self.norm_lv2(text_feats + self.ffn_lv(text_feats))

        # cross attend vision to language
        vis_feats2 = self.cross_vl(
            query=qv.transpose(0, 1),
            key=kt.transpose(0, 1),
            value=vt.transpose(0, 1),
            attn_mask=None,
            key_padding_mask=text_key_padding_mask  # (B, L)
        )[0].transpose(0, 1)
        vis_feats = vis_feats + self.dropout_vl(vis_feats2)
        vis_feats = self.norm_vl(vis_feats)
        vis_feats = self.norm_vl2(vis_feats + self.ffn_vl(vis_feats))

        return vis_feats, text_feats


class DeformableTransformerEncoderLayer(nn.Module):
    """TransformerEncoderLayer but additionaly add pos_embed in query."""

    def __init__(
        self,
        d_model,
        nhead,
        dim_feedforward=2048,
        dropout=0.1,
        activation="relu",
        n_levels=1,
        n_points=8,
    ):
        super().__init__()
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)

        # deformable attention
        self.deform_self_attn = MSDeformAttnPC(
            d_model=d_model,
            n_levels=n_levels,
            n_heads=nhead,
            n_points=n_points
        )

    def forward(
        self,
        src,
        pos,
        src_xyzs,
        reference_points,
        input_spatial_shapes,
    ):
        """
        Pass the input through the encoder layer (same as parent class).
        Args:
            src: (B, S, F)
            pos: (B, S, F)
            src_mask: the mask for the src sequence (optional)
            src_key_padding_mask: (B, S) mask for src keys per batch (optional)
            src_xyzs: (B, S, 3) (flattened pcs)
            reference_points: (B, N, L, 3) (unflattened pcs)
        Shape:
            see the docs in Transformer class.
        Return_shape: (S, B, F)
        """
        src2, aux_loss = self.deform_self_attn(
            query=src+pos,
            input_flatten=src,
            input_xyzs=src_xyzs,
            reference_points=reference_points, # for encoder, reference points are same as input xyzs
            input_spatial_shapes=input_spatial_shapes,
        )

        src = src + self.dropout1(src2)
        src = self.norm1(src)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)

        return src, aux_loss


class BiEncoderLayer(nn.Module):
    """Self->cross layer for both modalities."""

    def __init__(self, d_model=256, dropout=0.1, activation="relu", n_heads=8,
                 dim_feedforward=256):
        """Initialize layers, d_model is the encoder dimension."""
        super().__init__()

        # self attention in language
        self.self_attention_lang = nn.TransformerEncoderLayer(
            d_model=d_model,
            nhead=n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation
        )

        # self attention in vision
        self.self_attention_visual = DeformableTransformerEncoderLayer(
            d_model=d_model,
            nhead=n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation
        )

        # cross attention in language and vision
        self.cross_layer = CrossAttentionLayer(d_model, dropout, n_heads)

    def forward(self, vis_feats, pos_feats, padding_mask, text_feats,
                text_padding_mask, src_xyzs, reference_points,
                input_spatial_shapes):
        """
        Forward pass, feats (B, N, F), masks (B, N), diff N for V/L.
        src_xyzs: (B, S, 3) (flattened pcs)
        reference_points: (B, N, L, 3) (unflattened pcs)
        input_spatial_shapes: [N1, N2....., NL] num points in layer l
        """
        # Self attention for image
        vis_feats, aux_loss_vsa = self.self_attention_visual(
            vis_feats,
            pos_feats,
            src_xyzs,
            reference_points,
            input_spatial_shapes,
        )

        # Self attention for language
        text_feats = self.self_attention_lang(
            text_feats.transpose(0, 1),
            src_key_padding_mask=text_padding_mask
        ).transpose(0, 1)

        # Cross attention
        vis_feats, text_feats = self.cross_layer(
            vis_feats=vis_feats,
            vis_key_padding_mask=padding_mask,
            text_feats=text_feats,
            text_key_padding_mask=text_padding_mask,
            pos_feats=pos_feats
        )

        return vis_feats, text_feats, aux_loss_vsa


class BiEncoder(nn.Module):
    """Encode jointly language and vision."""

    def __init__(self, bi_layer, num_layers):
        """Pass initialized BiEncoderLayer and number of such layers."""
        super().__init__()
        self.layers = _get_clones(bi_layer, num_layers)
        self.num_layers = num_layers

    def forward(self, vis_feats, pos_feats, padding_mask, text_feats,
                text_padding_mask, src_xyzs, reference_points,
                input_spatial_shapes):
        """Forward pass, feats (B, N, F), masks (B, N), diff N for V/L."""
        aux_loss_vsa_list = []
        for _, layer in enumerate(self.layers):
            vis_feats, text_feats, aux_loss_vsa_l = layer(
                vis_feats,
                pos_feats,
                padding_mask,
                text_feats,
                text_padding_mask,
                src_xyzs,
                reference_points,
                input_spatial_shapes,
            )
            aux_loss_vsa_list.append(aux_loss_vsa_l)
        return vis_feats, text_feats, aux_loss_vsa_list


class PosTransformerEncoderLayer(nn.TransformerEncoderLayer):
    """TransformerEncoderLayer but additionaly add pos_embed in query."""

    def __init__(self, d_model, nhead, dim_feedforward, dropout, activation):
        """Intialize same as parent class."""
        super().__init__(d_model, nhead, dim_feedforward, dropout, activation)

    def forward(self, src, pos, src_mask=None, src_key_padding_mask=None):
        """
        Pass the input through the encoder layer (same as parent class).

        Args:
            src: (S, B, F)
            src_mask: the mask for the src sequence (optional)
            src_key_padding_mask: (B, S) mask for src keys per batch (optional)

        Shape:
            see the docs in Transformer class.
        Return_shape: (S, B, F)
        """
        src2 = self.self_attn(
            src + pos, src, src,
            attn_mask=src_mask,
            key_padding_mask=src_key_padding_mask
        )[0]
        src = src + self.dropout1(src2)
        src = self.norm1(src)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)
        return src


class BiDecoderLayer(nn.Module):
    """Self->cross_l->cross_v layer for proposals."""

    def __init__(self, d_model, n_heads, dim_feedforward=2048, dropout=0.1,
                 activation="relu",
                 self_position_embedding='loc_learned',
                 n_levels=1, n_points=8):
        """Initialize layers, d_model is the encoder dimension."""
        super().__init__()

        # Self attention
        self.self_attention = PosTransformerEncoderLayer(
            d_model=d_model,
            nhead=n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation
        )

        # Cross attention in language
        self.cross_l = nn.MultiheadAttention(
            d_model, n_heads, dropout=dropout
        )
        self.dropout_l = nn.Dropout(dropout)
        self.norm_l = nn.LayerNorm(d_model)
        self.ffn_l = nn.Sequential(
            nn.Linear(d_model, 1024),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(1024, d_model),
            nn.Dropout(dropout)
        )
        self.norm_l2 = nn.LayerNorm(d_model)

        # Cross attention in vision
        self.cross_v = MSDeformAttnPC(
            d_model=d_model,
            n_levels=n_levels,
            n_heads=n_heads,
            n_points=n_points
        )
        self.dropout_v = nn.Dropout(dropout)
        self.norm_v = nn.LayerNorm(d_model)
        self.ffn_v = deepcopy(self.ffn_l)
        self.norm_v2 = nn.LayerNorm(d_model)

        # Positional embeddings
        if self_position_embedding == 'xyz_learned':
            self.self_posembed = PositionEmbeddingLearned(3, d_model)
        elif self_position_embedding == 'loc_learned':
            self.self_posembed = PositionEmbeddingLearned(6, d_model)
        else:
            self.self_posembed = None

    def forward(self, query, vis_feats, lang_feats, query_pos,
                padding_mask, text_key_padding_mask, src_xyzs,
                reference_points, input_spatial_shapes):
        """
        Forward pass.

        Args:
            query: (B, N, F)
            vis_feats: (B, V, F)
            lang_feats: (B, L, F)
            query_pos: (B, N, 3or6)
            padding_mask: (B, V)
            text_key_padding_mask: (B, L)
            src_xyzs: (B, S, 3) (flattened pcs)
            reference_points: (B, N, L, 3) (unflattened pcs)
            input_spatial_shapes: [N1, N2....., NL] num points in layer l
        Returns:
            query: (B, N, F)
        """
        # NxCxP to PxNxC
        if self.self_posembed is not None:
            query_pos = self.self_posembed(query_pos)
            query_pos = query_pos.transpose(1, 2).contiguous()
        else:
            query_pos = torch.zeros_like(query, device=query.device)

        # Self attention
        query = self.self_attention(
            query.transpose(0, 1),
            query_pos.transpose(0, 1)
        ).transpose(0, 1)

        # Cross attend to language
        query2 = self.cross_l(
            query=(query + query_pos).transpose(0, 1),
            key=lang_feats.transpose(0, 1),
            value=lang_feats.transpose(0, 1),
            attn_mask=None,
            key_padding_mask=text_key_padding_mask  # (B, L)
        )[0].transpose(0, 1)
        query = query + self.dropout_l(query2)
        query = self.norm_l(query)
        query = self.norm_l2(query + self.ffn_l(query))

        # Cross attend to vision
        query2, aux_loss_vca = self.cross_v(
            query=(query + query_pos),
            input_flatten=vis_feats,
            input_xyzs=src_xyzs,
            reference_points=reference_points,
            input_spatial_shapes=input_spatial_shapes,
        )
        query = query + self.dropout_v(query2)
        query = self.norm_v(query)
        query = self.norm_v2(query + self.ffn_v(query))

        return query, aux_loss_vca


def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    if activation == "glu":
        return F.glu
    raise RuntimeError(F"activation should be relu/gelu, not {activation}.")



if __name__ == "__main__":
    net = BiDecoderLayer(256, 8)
    outs = net(
        torch.rand(2, 16, 256),
        torch.rand(2, 6, 256),
        torch.rand(2, 8, 256),
        torch.rand(2, 16, 6),
        torch.zeros(2, 6),
        torch.zeros(2, 8)
    )
    print(outs.shape)
